/*****************************************************************************/
/*                                                                           */
/* UNIT: NTL1_Abs_Spec  (Level 1 library routine)                            */
/*                                                                           */
/* Author: Nikola Stojanovic                                                 */
/*                                                                           */
/* Revision:    17 DEC 94   Version 1.0                                      */
/*              31 OCT 95   Version 2.0                                      */
/*                                                                           */
/* Function:                                                                 */
/*                                                                           */
/*   Procedure determines new offset (relative to the new origin) of an      */
/* absolute position in the alignment; structures containing data about the  */
/* origins and the alignment are provided as input parameters; returns the   */
/* standard error structure, NULL if there were no errors                    */
/*                                                                           */
/* NOTE: If there is no offset with respect to the given origin for the      */
/*       absolute position in the alignment, the list of offsets for return  */
/*       is empty (NULL), and NULL is returned for "no error", providing     */
/*       that there were no other problems during the evaluation (however,   */
/*       if the sequence containing the "target" offset is not present in    */
/*       the alignment at all, it is considered to be a problem)             */
/*                                                                           */
/*****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ntl1.h"

/*****************************************************************************/
/*                                                                           */
/* Definitions section                                                       */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/* Definitions of local constants of the unit                                */
/*****************************************************************************/


/* Limits for predefined buffer sizes                                        */

#define ERR_MSGLIMIT             128


/*****************************************************************************/
/* Prototypes of all locally used functions of this unit                     */
/*****************************************************************************/


errind NTL1_AS_Assemble_Error (int severity, int code,
                                char *comment, int description);


/*****************************************************************************/
/* Definitions of global (static) variables of the unit                      */
/*****************************************************************************/


static char Error_Message [ERR_MSGLIMIT]; /* Temporary buffer, error passing */

/*****************************************************************************/
/*                                                                           */
/* Code section                                                              */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/* PROCEDURE: NTL1_Abs_Spec                                                  */
/*                                                                           */
/* Central procedure for the unit; returns NULL if there were no errors in   */
/*   determining the corresponding position, error diagnostics structure     */
/*   otherwise                                                               */

errind NTL1_Abs_Spec (long int position, char *origin,
                      orgref_ptr origins, header_ptr file_data,
                      align_ptr alignment, appint_ptr *result)
{
 char *owner_seq; orgref_ptr org; line_ptr tracker; appint_ptr new_hit;
 long int new_pos, counter, at_abs, corresponding,
          trail_align, trail_seq, next_align, next_seq,
          trailer_absolute, trailer_offset;
 int count_found, index;

 /* Determine the "owner" sequnce(s) for "old" and "new" origin site         */

 owner_seq = NULL; new_pos = 0;
 for (org = origins; (org != NULL) && (owner_seq == NULL); org = org -> next)
  if (!strcmp (org -> site_name, origin)) {
   owner_seq = NTL0_strsave (org -> in_species); new_pos = org -> location;
  }
 if (owner_seq == NULL) {
  sprintf (Error_Message, "Site <%s> not found in origins file", origin);
  return NTL1_AS_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                 Error_Message, 1);
 }
 /* Check whether the sequence containing the "target" origin is present in  */
 /*   the alignment at all - return error if not                             */

 count_found = 0; index = 0;
 while ((count_found == 0) && (index < file_data -> dimension)) {
  if (!strcmp (((file_data -> sequences) [index]).seq_name, owner_seq)) {
   count_found++;
  }
  else index++;
 }
 if (count_found == 0) {
  sprintf (Error_Message,
           "Sequence <%s> of origin <%s> not found in the alignment",
           owner_seq, origin);
  return NTL1_AS_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                 Error_Message, 2);
 }
 /* Locate the target sequence in alignment that has character at position   */
 /* Check whether the located target sequence is unique within alignment     */

 *result = NULL;
 count_found = 0; at_abs = 0; trailer_absolute = -1; trailer_offset = 0;
 for (index = 0; index < file_data -> dimension; index++) {
  if (!strcmp (((file_data -> sequences) [index]).seq_name, owner_seq)) {
   trail_align = -1; trail_seq = (alignment -> begin) [index] - 1;
   counter = 0; corresponding = (alignment -> begin) [index];
   for (tracker = (alignment -> lines) [index];
        (tracker != NULL) && (counter <= position);
        tracker = tracker -> next) {
    if (tracker -> code == GAP_STRETCH) counter += tracker -> length;
    else if (tracker -> code == TEXT_STRETCH) {
     if (counter + tracker -> length > position) {
      count_found++; at_abs = corresponding + position - counter;
      new_hit = (appint_ptr) NTL0_ckalloc (sizeof (AppInt_Struct));
      if (at_abs >= new_pos) new_hit -> value = at_abs - new_pos + 1;
      else new_hit -> value = at_abs - new_pos;
      new_hit -> correction = 0;
      new_hit -> next = *result; *result = new_hit;
     }
     else {                        /* Continue counting through the sequence */
      corresponding += tracker -> length;

      /* Keep track of position of the end of the last strip of text seen    */
      /*   in this sequence, in the alignment and in seq. absolute pos.      */

      trail_align = counter + tracker -> length - 1;
      trail_seq = trail_seq + tracker -> length;
     }
     counter += tracker -> length;
    }
   }
   /* If no corresponding position has been found yet, find the beginning    */
   /*   of the next text piece in the sequence (if any), its offset from     */
   /*   the beginning of the alignment and its position within its sequence  */
   /* Next, compare the recorded value of the smallest distance between the  */
   /*   neighbouring edges and the position with new distances and update    */
   /*   values, if necessary: value of the absolute alignment position of    */
   /*   the nearest neighbouring edge and its offset within its sequence     */

   if (count_found == 0) {
    if (tracker == NULL) { next_align = -1; next_seq = -1; }
    else if (counter < alignment -> size) {
     next_align = counter;
     if (trail_align == -1) next_seq = (alignment -> begin) [index];
     else next_seq = trail_seq + 1;
    }
    if (trail_align != -1) {
     if (next_align != -1) {
      if (position - trail_align > next_align - position) {
       if (trailer_absolute < 0) {
        trailer_absolute = next_align; trailer_offset = next_seq;
       }
       else {
        if (NTL0_labs (position - trailer_absolute) > next_align - position) {
         trailer_absolute = next_align; trailer_offset = next_seq;
        }
       }
      }
      else {
       if (trailer_absolute < 0) {
        trailer_absolute = trail_align; trailer_offset = trail_seq;
       }
       else {
        if (NTL0_labs (position - trailer_absolute) > position - trail_align) {
         trailer_absolute = trail_align; trailer_offset = trail_seq;
        }
       }
      } 
     }
     else {           /* There is no next neighbouring position in alignment */
      if (trailer_absolute < 0) {
       trailer_absolute = trail_align; trailer_offset = trail_seq;
      }
      else {
       if (NTL0_labs (position - trailer_absolute) > position - trail_align) {
        trailer_absolute = trail_align; trailer_offset = trail_seq;
       }
      }
     }
    }
    else {
     if (next_align != -1) {
      if (trailer_absolute < 0) {
       trailer_absolute = next_align; trailer_offset = next_seq;
      }
      else {
       if (NTL0_labs (position - trailer_absolute) > next_align - position) {
        trailer_absolute = next_align; trailer_offset = next_seq;
       }
      }
     }
    }
   }
  }
 }
 if (count_found == 0) {
  if (trailer_absolute >= 0) {
  
   /* Although there was no direct correspondent to the given position in    */
   /*   the alignment, there was a neighbouring position found; return the   */
   /*   neighbours offset and the size of the gap dividing the "right"       */
   /*   corresponding position and the selected neighbour                    */

   new_hit = (appint_ptr) NTL0_ckalloc (sizeof (AppInt_Struct));
   if (trailer_offset >= new_pos)
    new_hit -> value = trailer_offset - new_pos + 1;
   else new_hit -> value = trailer_offset - new_pos;
   if (position < trailer_absolute)
    new_hit -> correction = trailer_absolute - position;
   else new_hit -> correction = trailer_absolute - position;
   new_hit -> next = *result; *result = new_hit;
  }
 }
 /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
 /* if (*result == NULL) {                                                   */
 /*  sprintf (Error_Message,                                                 */
 /*           "No match for <%ld> in alignment, with respect to <%s>",       */
 /*           position, origin);                                             */
 /*  return NTL1_AS_Assemble_Error (USER_ERROR, ERR_NO_VALUE,                */
 /*                                 Error_Message, 3);                       */
 /* }                                                                        */
 /* else return NULL;                                                        */
 /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

 return NULL;
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL1_AS_Assemble_Error                                         */
/*                                                                           */
/* Service procedure for assembling and returnning an error report, based on */
/*   the values of the input parameters; returns the record with the report  */

errind NTL1_AS_Assemble_Error (int severity, int code,
                               char *comment, int description)
{
 char *report; errind assembled;

 report = (char *) NTL0_ckalloc (
                   (strlen (comment) + strlen ("_Absolute_to_Species: ") + 1) *
                                                                sizeof (char));
 sprintf (report, "_Absolute_to_Species: %s", comment);
 assembled = NTL1_Error_Record (severity, code, report, description);
 free (report); return assembled;
}

