/*****************************************************************************/
/*                                                                           */
/* UNIT: NTL2_Load_ColorUnderlays (Level 2 library routine)                  */
/*                                                                           */
/* Author: Liliana Florea                                                    */
/*                                                                           */
/* Revision:    06 JUN 06   Version 1.0                                      */
/*                                                                           */
/* Function:                                                                 */
/*                                                                           */
/* This procedure loads a file containing specification of color underlays.  */
/* Method: the procedure receives the name of the underlay file, retrieves   */
/* and stores the data into a record list. It returns NULL if everything     */
/* was OK.                                                                   */ 
/*                                                                           */
/*                                                                           */
/* Expected format of the color underlay file:                               */
/*                                                                           */
/* - Any line can contain a comment, starting with '#' symbol - the rest of  */
/*   the line is then considered a comment, and ignored by the procedure.    */
/* - Any data line contains:                                                 */
/*                                                                           */
/*    s_1   s_2  colortxt                                                    */
/*                                                                           */
/*    s_1 = starting position of the region represented by the line          */
/*    s_2 = ending position of the region represented by the line            */
/*    colortxt = name of the color used for underlay                         */
/*                                                                           */ 
/*****************************************************************************/

#include <stdio.h> 
#include <stdlib.h> 
#include <ctype.h>
#include <string.h> 
#include "ntl2.h" 

/*****************************************************************************/ 
/*                                                                           */ 
/* Definitions section                                                       */ 
/*                                                                           */ 
/*****************************************************************************/ 


/*****************************************************************************/ 
/* Definitions of local constants of the unit                                */ 
/*****************************************************************************/ 


#define ERR_MSGLIMIT             128    /* size limit for error message */
#define IN_LIMIT                 200    /* size limit for text buffer */

/*****************************************************************************/
/* Prototypes of all locally used functions of this unit                     */
/*****************************************************************************/

static errind NTL2_LC_Read_ColorUnderlay_File (FILE **fileptr, 
                               colorunderlay_ptr *records);
static errind NTL2_Parse_ColorUnderlay_Line(char *In_Buff, 
                        colorunderlay_ptr *new_cunderlay);
static errind NTL2_LC_Assemble_Error (int severity, int code,
                               char *comment, int description);
                               
/*****************************************************************************/
/* Definitions of global (static) variables of the unit                      */
/*****************************************************************************/

static char In_Buff [IN_LIMIT]; /* Global input line buffer, for convenience */
static char Error_Message [ERR_MSGLIMIT]; /* Temporary buffer, error passing */

/*****************************************************************************/
/*                                                                           */
/* Code section                                                              */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_Load_ColorUnderlays                                       */
/*                                                                           */
/* Main (interface) procedure of this unit                                   */

errind NTL2_Load_ColorUnderlays (char *file_name, colorunderlay_ptr *records)
{
  FILE *fp; errind erret;

  *records = NULL;

  if ((fp = fopen (file_name, "r")) == NULL) {     /* Error condition */

     sprintf (Error_Message, "File <%s> does not exist)`", file_name);
     return NTL2_LC_Assemble_Error (USER_ERROR, ERR_NO_FILE, Error_Message, 0);
  } else {
     /* read-in the records */
     erret = NTL2_LC_Read_ColorUnderlay_File (&fp, records); 
     if ((erret != NULL) && (erret -> kind != WARNING)) { 
         /* free records list - it won't be needed  */
         *records = NTL1_Destroy_ColorUnderlay_List (*records);
     }
     fclose (fp); 
     return erret;
  }
}

    
/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_LC_Read_ColorUnderlay_File                                */
/*                                                                           */
/* Top level procedure for reading the color underlay file contents;         */
/*   loops for every line collecting the data (one record per line);         */
/*   receives the file pointer and (already initialized) list of records to  */
/*   be filled, returns the assembled error record, NULL if there were no    */
/*   errors in loading                                                       */

static errind NTL2_LC_Read_ColorUnderlay_File (FILE **fptr, colorunderlay_ptr *records)
{
  colorunderlay_ptr new_cunderlay;
  char *s;
  errind erret;
  
  /* Get the file identification first, to verify the data format */
   
  while (fgets(In_Buff, IN_LIMIT, *fptr) && In_Buff[0]!='\0') {
     if (In_Buff[0]=='#') continue;                         /* skip comments */
     s = In_Buff;
     while (isspace(*s) && *s!='\n') s++;
     if (*s=='\n') continue;                             /* skip blank lines */
     
     erret = NTL2_Parse_ColorUnderlay_Line(In_Buff, &new_cunderlay);

     if (erret && (erret->kind != WARNING)) 
        return erret;   /* will close the file pointer in the parent routine */
     else {
        new_cunderlay->next = *records;        /* add new record to the list */
        *records = new_cunderlay;
     }
  }

  return NULL;  
}

/*****************************************************************************/
/* NTL2_Parse_ColorUnderlay_Line()                                           */
/*                                                                           */
/* Parse a line in the color underlay file;                                  */
/* Line format: n1 n2 colorname                                              */
/* Return error message, NULL is OK.                                         */

static errind NTL2_Parse_ColorUnderlay_Line(char *In_Buff, colorunderlay_ptr *new_cunderlay)
{
  char *s, *q, *aux, *colorname;
  int start_pos, end_pos;

  s = In_Buff;
  while (isspace(*s) && *s!='\n') s++;
  if (*s=='\n')   
     return NTL2_LC_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                    "Illegal line in color underlay file", 1);
  
  /* check if the line format is valid: read the first column */  
  q = s;
  while (isdigit(*s)) s++;            /* negative positions are not accepted */
  if (q==s) 
      return NTL2_LC_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                     "Illegal line in color underlay file", 2);
  sscanf(q, "%d", &start_pos);

  /* read the second column */
  q = s;
  while (isspace(*s)) s++; 
  if (q==s)
      return NTL2_LC_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                     "Illegal line in color underlay file", 3);
  q = s;
  while (isdigit(*s)) s++; 
  if (q==s)
      return NTL2_LC_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                     "Illegal line in color underlay file", 4);
  sscanf(q, "%d", &end_pos);

  /* read the third column */
  q = s;
  while (isspace(*s)) s++;
  if (q==s)
      return NTL2_LC_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                     "Illegal line in color underlay file", 5);
  q = s;                             
  while (!isspace(*s)) s++;
  if (q==s)
      return NTL2_LC_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                     "Illegal line in color underlay file", 6);
  aux = s;
  while (isspace(*s) && *s!='\n') s++;
  if (*s!='\n') 
     return NTL2_LC_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                     "Illegal line in color underlay file", 7);
  /* save the color name */
  *aux = '\0';
  colorname = q; 

  *new_cunderlay = (colorunderlay_ptr) NTL0_ckalloc(sizeof(ColorUnderlay_Struct));
  (*new_cunderlay)->start = start_pos;
  (*new_cunderlay)->stop = end_pos;
  (*new_cunderlay)->color_name = NTL0_strsave (colorname);
  (*new_cunderlay)->next = NULL;

  return NULL;
}

/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_LC_Assemble_Error                                         */
/*                                                                           */
/* Service procedure for assembling and returnning an error report, based on */
/*   the values of the input parameters; returns the record with the report  */

errind NTL2_LC_Assemble_Error (int severity, int code,
                               char *comment, int description)
{                              
 char *report; errind assembled;
 
 report = (char *) NTL0_ckalloc (
       (strlen (comment) + strlen ("_Load_ColorUnderlay_File: ") + 1) * sizeof (char));
 sprintf (report, "_Load_ColorUnderlay_File: %s", comment);
 assembled = NTL1_Error_Record (severity, code, report, description);
 free (report); return assembled;
}
