/*****************************************************************************/
/*                                                                           */
/* UNIT: NTL2_Load_Origins  (Level 2 library routine)                        */
/*                                                                           */
/* Author: Nikola Stojanovic                                                 */
/*                                                                           */
/* Revision:    06 JUL 94   Version 1.0                                      */
/*              15 OCT 94   Version 2.0                                      */
/*              10 JUL 95   Version 2.1                                      */
/*                                                                           */
/* Function:                                                                 */
/*                                                                           */
/*   Procedure loads the "origin structure" (data structure containing       */
/* information about the known origin sites in sequences) from the specified */
/* file containing origins information in the "database tuples" format.      */
/*   Procedure receives the name of the file containing origins data, fills  */
/* the pointer (reference parameter) to the begining of the assembled list   */
/* of origins; returns the error structure, NULL if everything was OK        */
/*                                                                           */
/*                                                                           */
/* Expected format of the "origins" file:                                    */
/*                                                                           */
/* - First non-white-space sequence of symbols in the file must be:          */
/*                                                                           */
/*   #:TUPLES:                                                               */
/*                                                                           */
/* -                                                                         */
/*                                                                           */
/*****************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ntl2.h"

/*****************************************************************************/
/*                                                                           */
/* Definitions section                                                       */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/* Definitions of local constants of the unit                                */
/*****************************************************************************/


/* Limit for the length of an error message that can be assembled            */

#define ERR_MSGLIMIT             128

/* Information is read from database files by supplementary buffer of size:  */

#define IN_LIMIT                 512

#define LNUM_SIZE                8

/* Values of "tokens" found in origins file                                  */

#define FTOK_EOF                 0
#define FTOK_ITEM                2
#define FTOK_SPECIES             4
#define FTOK_ORIGIN              5
#define FTOK_SPEC                6

#define FTOK_SITE                11
#define FTOK_OFFSET              12



/*****************************************************************************/
/* Prototypes of all locally used functions of this unit                     */
/*****************************************************************************/


errind NTL2_LO_Read_Origins (FILE **orig_file, orgref_ptr *origins);
errind NTL2_LO_Next_Origin (FILE **orig_file,
                            orgref_ptr origin_rec, orgref_ptr *new_record);
errind NTL2_LO_Origin_Token (FILE **orig_file, int *next_tok, char **str);
errind NTL2_LO_Fill_Origin (orgref_ptr origin_rec, int next_tok, char **str);
errind NTL2_LO_Assemble_Error (int severity, int code,
                               char *comment, int description);


/*****************************************************************************/
/* Definitions of global (static) variables of the unit                      */
/*****************************************************************************/


static char In_Buff [IN_LIMIT]; /* Global input line buffer, for convenience */

static char Error_Message [ERR_MSGLIMIT]; /* Temporary buffer, error passing */

/*****************************************************************************/
/*                                                                           */
/* Code section                                                              */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_Load_Origins                                              */
/*                                                                           */
/* Main (interface) procedure of this unit                                   */

errind NTL2_Load_Origins (char *file_name, orgref_ptr *origins)
{
 FILE *orig_file; char ch; errind erret; orgref_ptr sscan, tscan;

 /* Initialize the list of origins to return to "empty", before actions      */

 *origins = NULL;

 /* Check whether the file with given file name exists and open it for read  */

 if ((orig_file = fopen (file_name, "r")) == NULL) {      /* Error condition */
  sprintf (Error_Message, "File <%s> does not exist", file_name);
  return NTL2_LO_Assemble_Error (USER_ERROR, ERR_NO_FILE, Error_Message, 1);
 }
 else {                           /* There is a file with the specified name */

  /* Check the file format by means of the recorded header code              */

  do {                        /* Eliminate the initial "white space" symbols */
      ch = fgetc (orig_file);
  } while ((ch == ' ') || (ch == '\n') || (ch == '\t'));

  if (ch != '#') return  NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                                 "Illegal file format", 2);
  ch = fgetc (orig_file); if (ch != ':')
   return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                  "Illegal file format", 3);
  ch = fgetc (orig_file); if ((ch != 't') && (ch != 'T'))
   return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                  "Illegal file format", 4);
  ch = fgetc (orig_file); if ((ch != 'u') && (ch != 'U'))
   return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                  "Illegal file format", 5);
  ch = fgetc (orig_file); if ((ch != 'p') && (ch != 'P'))
   return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                  "Illegal file format", 6);
  ch = fgetc (orig_file); if ((ch != 'l') && (ch != 'L'))
   return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                  "Illegal file format", 7);
  ch = fgetc (orig_file); if ((ch != 'e') && (ch != 'E'))
   return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                  "Illegal file format", 8);
  ch = fgetc (orig_file); if ((ch != 's') && (ch != 'S'))
   return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                  "Illegal file format", 9);
  ch = fgetc (orig_file); if (ch != ':')
   return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                  "Illegal file format", 10);

  /* Now load the origins from the file and close the file when done         */

  erret = NTL2_LO_Read_Origins (&orig_file, origins);       /* Read the file */
  if ((erret != NULL) && (erret -> kind != WARNING)) {  /* Error status ret. */
   /* Release the assembled part of the origins list - it won't be needed    */

   sscan = *origins; while (sscan != NULL) {
    tscan = sscan; sscan = sscan -> next; free (tscan);
   }
   *origins = NULL;
  }
  fclose (orig_file); return erret;
 }
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_LO_Read_Origins                                           */
/*                                                                           */
/* Top level procedure for reading the origins file contents; loops for each */
/*   new origin record in the file (and accounts for field contents          */
/*   inherritance, as well) and loads it via lower-level procedures.         */
/* Procedure receives the file pointer and (already initialized) list of     */
/*   origin sites; returns the assembled error record, NULL if there was no  */
/*   error in loading                                                        */

errind NTL2_LO_Read_Origins (FILE **orig_file, orgref_ptr *origins)
{
 orgref_ptr origin_rec, new_origin; errind erret;

 /* Allocate a record to store the new loaded origin before passing it to    */
 /*   procedure: since there is no attribute to inherit, initialize contents */

 origin_rec = (orgref_ptr) NTL0_ckalloc (sizeof (OrgRef_Struct));
 origin_rec -> in_species = NULL;
 origin_rec -> species = NULL;
 origin_rec -> site_name = NULL;
 origin_rec -> location = 0;
 origin_rec -> next = NULL;

 /* Loop to get origins until there are no more in the file                  */

 while (((erret = NTL2_LO_Next_Origin (orig_file,
                                       origin_rec, &new_origin)) == NULL) &&
         (new_origin != NULL)) {

  /* Allocate the next record and prepare to copy previous contents for inhr.*/

  origin_rec = (orgref_ptr) NTL0_ckalloc (sizeof (OrgRef_Struct));
  if (new_origin -> in_species != NULL)
   origin_rec -> in_species = NTL0_strsave (new_origin -> in_species);
  else origin_rec -> in_species = NULL;
  if (new_origin -> species != NULL)
   origin_rec -> species = NTL0_strsave (new_origin -> species);
  else origin_rec -> species = NULL;
  if (new_origin -> site_name != NULL)
   origin_rec -> site_name = NTL0_strsave (new_origin -> site_name);
  else origin_rec -> site_name = NULL;
  origin_rec -> location = new_origin -> location;
  origin_rec -> next = NULL;

  /* Now link the new origin record to the list of already loaded origins    */

  new_origin -> next = *origins; *origins = new_origin;
 }
 /* At the end of the loop no origin was found, so release the preallocated  */

 if (origin_rec -> in_species != NULL) free (origin_rec -> in_species);
 if (origin_rec -> species != NULL) free (origin_rec -> species);
 if (origin_rec -> site_name != NULL) free (origin_rec -> site_name);
 free (origin_rec);

 return erret;      /* Return whatever was the error status from lower level */
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_LO_Next_Origin                                            */
/*                                                                           */
/* Procedure extracts information about one origin at a time from the file   */
/*   and loads it into a new origin structure, which coincides with the      */
/*   structure passed in the procedure (if there were no further origin      */
/*   block information in the file, that value returns as NULL); returns the */
/*   "standard" error structure, NULL if there were no errors in loading     */
/*   the structure                                                           */

errind NTL2_LO_Next_Origin (FILE **orig_file,
                            orgref_ptr origin_rec, orgref_ptr *new_record)
{
 int next_tok; char *str; errind status;

 *new_record = NULL;    /* Start with assumption that there is no new origin */

 /* Get the next relevant token from the file: if it indicates a start of    */
 /*   entirely new origins block, reset inherited fields in reserved record  */

 if ((status = NTL2_LO_Origin_Token (orig_file, &next_tok, &str)) != NULL) {
  return status;
 }
 else if (next_tok == FTOK_ORIGIN) {
  if (origin_rec -> in_species != NULL) free (origin_rec -> in_species);
  if (origin_rec -> species != NULL) free (origin_rec -> species);
  if (origin_rec -> site_name != NULL) free (origin_rec -> site_name);
  origin_rec -> in_species = origin_rec -> site_name = NULL;
  origin_rec -> species = NULL;
  origin_rec -> location = 0;
 }
 else if (next_tok == FTOK_ITEM) { /* There was supposed to be some contents */
  *new_record = origin_rec;
  return NTL2_LO_Assemble_Error (WARNING, ERR_FILE_FORMAT,
                                 "No contents in origin block", 11);
 }
 else if (next_tok == FTOK_EOF) return NULL;        /* EOF - no more origins */
 else if ((status = NTL2_LO_Fill_Origin (origin_rec, next_tok, &str)) != NULL)
  return status;                /* Origin contents or error in assembling it */

 /* At this point it is not possible that there won't be a new origin - fill */

 *new_record = origin_rec;   /* Whatever is now in old gets into the new one */

 /* Now proceed to get all the attributes of the origin that has been seen   */

 do {
     status = NTL2_LO_Origin_Token (orig_file, &next_tok, &str); /* Token... */
     if (status != NULL) return status;          /* Error in the file layout */
     else {
      switch (next_tok) {
       case FTOK_ORIGIN: { /* Already started - no other start before "item" */
        return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                       "Illegal file format - double origins",
                                       12);
       }
       case FTOK_ITEM: return NULL;        /* So full contents are collected */

       case FTOK_EOF: {      /* Any origin must complete ("item") before EOF */
        return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT, 
                                       "Illegal file format - unterminated",
                                       13);
       }
       default:
        if ((status = NTL2_LO_Fill_Origin (origin_rec, next_tok, &str))
                    != NULL) return status; /* Attribute or error assembling */
      }
     }
 } while (TRUE);      /* The endless loop will be broken when "item" returns */
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_LO_Origin_Token                                           */
/*                                                                           */
/* Procedure gets the next "token" from the origins file (or file containing */
/*   origins) plus the accompanying information in the form of a string, if  */
/*   any; returns the error structure, NULL if there was no error discovered */

errind NTL2_LO_Origin_Token (FILE **orig_file, int *next_tok, char **str)
{
 char ch; int accept_pos; int pat, reached, in_comment; errind status;

 /* Skip the heading white spaces and all comments possibly interleaved      */

 while (TRUE) {
  ch = ' '; while ((ch == ' ') || (ch == '\n') || (ch == '\t'))
   ch = fgetc (*orig_file);
  if (ch == EOF) { *str = NULL; *next_tok = FTOK_EOF; return NULL; }
  else if (ch == '#') {
   in_comment = TRUE; while (in_comment) {
    while ((ch != '\n') && (ch != EOF)) ch = fgetc (*orig_file);
    in_comment = FALSE;
    while ((ch == ' ') || (ch == '\n') || (ch == '\t'))
     ch = fgetc (*orig_file);
    if (ch == EOF) { *str = NULL; *next_tok = FTOK_EOF; return NULL; }
    else if (ch == '#') in_comment = TRUE;
   }
  }

  /* At this point some significant text has been seen - investigate         */

  accept_pos = 0;
  while ((accept_pos < IN_LIMIT) &&
         (ch != EOF) && (ch != ' ') && (ch != '\n') && (ch != '\t') &&
         (ch != ';') && (ch != '=')) {
   In_Buff [accept_pos++] = ch; ch = fgetc (*orig_file);
  }
  if (accept_pos >= IN_LIMIT) {                           /* Error condition */
   return NTL2_LO_Assemble_Error (USER_ERROR, ERR_BUFFER_OVERFLOW,
                                  "Text in file too long", 14);
  }
  else if (ch == EOF) {                       /* End-of-file prematurely hit */
   return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                  "Unexpected end-of-file", 15);
  }
  else {     /* Text up to first "white space" is placed in accepting buffer */

   In_Buff [accept_pos] = '\0';

   if ((!strcmp (In_Buff, "begin")) ||
       (!strcmp (In_Buff, "BEGIN"))) {          /* Beginning of some block */
    while ((ch == ' ') || (ch == '\n') || (ch == '\t'))
     ch = fgetc (*orig_file);
    accept_pos = 0;
    while ((accept_pos < IN_LIMIT) &&
           (ch != EOF) && (ch != ' ') && (ch != '\n') && (ch != '\t')) {
     In_Buff [accept_pos++] = ch; ch = fgetc (*orig_file);
    }
    if (accept_pos >= IN_LIMIT) {                         /* Error condition */
     return NTL2_LO_Assemble_Error (USER_ERROR, ERR_BUFFER_OVERFLOW,
                                    "Text in file too long", 16);
    }
    else if (ch == EOF) {                                 /* Error condition */
     return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                    "Unexpected end-of-file", 17);
    }
    else {       /* Name of the block collected in acceptance buffer - check */

     In_Buff [accept_pos] = '\0';
     if (!strcmp (In_Buff, "origin_info")) {
      *str = NULL; *next_tok = FTOK_ORIGIN;     /* Beginning of a new origin */
      return NULL;
     }
     else {    /* So this was beginning of something else - find next origin */

      reached = FALSE; while ((ch != EOF) && (!reached)) {
       while ((ch != EOF) && (ch != 'b') && (ch != 'B'))
        ch = fgetc (*orig_file);
       if (ch != EOF) {
        pat = TRUE;
        ch = fgetc (*orig_file);
        if ((ch != 'e') && (ch != 'E')) pat = FALSE;
        if (pat) {
         ch = fgetc (*orig_file);
         if ((ch != 'g') && (ch != 'G')) pat = FALSE;
        }
        if (pat) {
         ch = fgetc (*orig_file);
         if ((ch != 'i') && (ch != 'I')) pat = FALSE;
        }
        if (pat) {
         ch = fgetc (*orig_file);
         if ((ch != 'n') && (ch != 'N')) pat = FALSE;
        }
        if (pat) {
         ch = fgetc (*orig_file);
         if ((ch != ' ') && (ch != '\t') && (ch != '\n')) pat = FALSE;
         if (pat) {
          /* Eliminate the "white space" symbols between "begin" and ident.  */
          
          while ((ch == ' ') || (ch == '\t') || (ch == '\n'))
           ch = fgetc (*orig_file);

          /* Check if the identifier of what starts is "origin"              */
          
          if ((ch != 'o') && (ch != 'O')) pat = FALSE;
          if (pat) {
           ch = fgetc (*orig_file);
           if ((ch != 'r') && (ch != 'R')) pat = FALSE;
          }
          if (pat) {
           ch = fgetc (*orig_file);
           if ((ch != 'i') && (ch != 'I')) pat = FALSE;
          }
          if (pat) {
           ch = fgetc (*orig_file);
           if ((ch != 'g') && (ch != 'G')) pat = FALSE;
          }
          if (pat) {
           ch = fgetc (*orig_file);
           if ((ch != 'i') && (ch != 'I')) pat = FALSE;
          }
          if (pat) {
           ch = fgetc (*orig_file);
           if ((ch != 'n') && (ch != 'N'))  pat = FALSE;
          }
          if (pat) {
           ch = fgetc (*orig_file);
           if (ch != '_') pat = FALSE;
          }
          if (pat) {
           ch = fgetc (*orig_file);
           if ((ch != 'i') && (ch != 'I')) pat = FALSE;
          }
          if (pat) {
           ch = fgetc (*orig_file);
           if ((ch != 'n') && (ch != 'N')) pat = FALSE;
          }
          if (pat) {
           ch = fgetc (*orig_file);
           if ((ch != 'f') && (ch != 'F')) pat = FALSE;
          }
          if (pat) {
           ch = fgetc (*orig_file);
           if ((ch != 'o') && (ch != 'O')) pat = FALSE;
          }
          if (pat) {
           ch = fgetc (*orig_file);
           if ((ch == ' ') || (ch == '\t') || (ch == '\n')) reached = TRUE;
          }
         }
        }
       }
      }
      /* At this point it is either a beginning of new origin or EOF         */

      if (ch == EOF) { *str = NULL; *next_tok = FTOK_EOF; return NULL; }
      else { *str = NULL; *next_tok = FTOK_ORIGIN; return NULL; }
     }
    }
   }
   /* Since the text seen was not beginning of block, must be in some origin */

   else if (!strcmp (In_Buff, "name")) {
    while ((ch == ' ') || (ch == '\t') || (ch == '\n'))
     ch = fgetc (*orig_file);
    if (ch != '=') {                                      /* Error condition */
     return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                    "Illegal file format", 18);
    }
    else {                      /* So name specification is OK, get the name */
     ch = fgetc (*orig_file);
     while ((ch == ' ') || (ch == '\t') || (ch == '\n'))
      ch = fgetc (*orig_file);
     if ((status = NTL2_Assemble_String (orig_file, ch, str)) == NULL) {
      *next_tok = FTOK_SITE; return NULL;
     }
     else return status;
    }
   }

   else if (!strcmp (In_Buff, "species")) {
    while ((ch == ' ') || (ch == '\t') || (ch == '\n'))
     ch = fgetc (*orig_file);
    if (ch != '=') {                                      /* Error condition */
     return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                    "Illegal file format", 19);
    }
    else {
     ch = fgetc (*orig_file);
     while ((ch == ' ') || (ch == '\t') || (ch == '\n'))
      ch = fgetc (*orig_file);
     if ((status = NTL2_Assemble_String (orig_file, ch, str)) == NULL) {
      *next_tok = FTOK_SPEC; return NULL;
     }
    }
   }

   else if (!strcmp (In_Buff, "file")) {
    while ((ch == ' ') || (ch == '\t') || (ch == '\n'))
     ch = fgetc (*orig_file);
    if (ch != '=') {                                      /* Error condition */
     return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                    "Illegal file format", 19);
    }
    else {
     ch = fgetc (*orig_file);
     while ((ch == ' ') || (ch == '\t') || (ch == '\n'))
      ch = fgetc (*orig_file);
     if ((status = NTL2_Assemble_String (orig_file, ch, str)) == NULL) {
      *next_tok = FTOK_SPECIES; return NULL;
     }
    }
   }

   else if (!strcmp (In_Buff, "offset")) {
    while ((ch == ' ') || (ch == '\t') || (ch == '\n'))
     ch = fgetc (*orig_file);
    if (ch != '=') {                                      /* Error condition */
     return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                    "Illegal file format", 20);
    }
    else {                  /* Proceed to find what follows the "equal" sign */
     ch = fgetc (*orig_file);
     while ((ch == ' ') || (ch == '\t') || (ch == '\n'))
      ch = fgetc (*orig_file);
     if ((ch != '-') && ((ch < '0') || (ch > '9'))) {     /* Error condition */
      return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                     "Illegal number in file", 21);
     }
     else {
      In_Buff [0] = ch; accept_pos = 1; ch = '0';
      while ((accept_pos < LNUM_SIZE) && (ch >= '0') && (ch <= '9')) {
       ch = fgetc (*orig_file); In_Buff [accept_pos++] = ch;
      }
      if (accept_pos >= LNUM_SIZE) {                      /* Error condition */
       return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                      "Illegal number in file", 22);
      }
      else if ((accept_pos == 2) && (In_Buff [0] == '-')) {
       return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                      "Illegal number in file", 23);
      }
      else {
       while ((ch == ' ') || (ch == '\t')) /* || (ch == '\n')) */
        ch = fgetc (*orig_file);
       /* if (ch != ';') { */
       if (ch != '\n') {                                 /* Error condition */
        return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                       "Illegal file format", 24);
       }
       else {
        accept_pos--; In_Buff [accept_pos] = '\0';
        *str = NTL0_strsave (In_Buff); *next_tok = FTOK_OFFSET; return NULL;
       }
      }
     }
    }
   }

   else if ((!strcmp (In_Buff, "end")) ||
            (!strcmp (In_Buff, "END"))) {       /* The end of current origin */
    while ((ch == ' ') || (ch == '\t') || (ch == '\n'))
     ch = fgetc (*orig_file);
    accept_pos = 0;
    while ((accept_pos < IN_LIMIT) &&
           (ch != EOF) && (ch != ' ') && (ch != '\n') && (ch != '\t')) {
     In_Buff [accept_pos++] = ch; ch = fgetc (*orig_file);
    }
    if (accept_pos >= IN_LIMIT) {                         /* Error condition */
     return NTL2_LO_Assemble_Error (USER_ERROR, ERR_BUFFER_OVERFLOW,
                                    "Text in file too long", 25);
    }
    else if (ch == EOF) {                                 /* Error condition */
     return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                    "Unexpected end-of-file", 26);
    }
    else {       /* Name of the block collected in acceptance buffer - check */

     In_Buff [accept_pos] = '\0';
     if (!strcmp (In_Buff, "origin_info")) {
      *str = NULL; *next_tok = FTOK_ITEM; return NULL;
     }
     return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                     "Illegal conserved block conclusion", 27);
    }
   }
   else {                                                 /* Error condition */
    return NTL2_LO_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                   "Illegal token for origin in file", 28);
   }
  }
 }
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_LO_Fill_Origin                                            */
/*                                                                           */
/* Procedure fills single received origin record with information passed to  */
/*   it, which depends on the token that caused the record change; returns   */
/*   NULL if everything was OK, or indication of problem in the program if   */
/*   an unknown type of token has been passed to it                          */

errind NTL2_LO_Fill_Origin (orgref_ptr origin_rec, int next_tok, char **str)
{
 switch (next_tok) {
  case FTOK_SITE: { origin_rec -> site_name = *str; *str = NULL; return NULL; }
  case FTOK_SPECIES: {
   origin_rec -> in_species = *str; *str = NULL; return NULL;
  }
  case FTOK_SPEC: {
   origin_rec -> species = *str; *str = NULL; return NULL;
  }
  case FTOK_OFFSET: { origin_rec -> location = atol (*str); return NULL; }
  default: {
   return NTL2_LO_Assemble_Error (FATAL_ERROR, ERR_CODE_PROBLEM,
                                  "Illegal token passed to assembling", 29);
  }
 }
}
    

/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_LO_Assemble_Error                                         */
/*                                                                           */
/* Service procedure for assembling and returnning an error report, based on */
/*   the values of the input parameters; returns the record with the report  */

errind NTL2_LO_Assemble_Error (int severity, int code,
                               char *comment, int description)
{
 char *report; errind assembled;

 report = (char *) NTL0_ckalloc (
                     (strlen (comment) + strlen ("_Load_Origins: ") + 1) *
                                                                sizeof (char));
 sprintf (report, "_Load_Origins: %s", comment);
 assembled = NTL1_Error_Record (severity, code, report, description);
 free (report); return assembled;
}

