/*****************************************************************************/
/*                                                                           */
/* UNIT: NTL2_Load_Sequence  (Level 2 library routine)                       */
/*                                                                           */
/* Author: Webb Miller                                                       */
/*         Nikola Stojanovic                                                 */
/*                                                                           */
/* Revision:    26 JUL 94   Version 1.0                                      */
/*              08 JUL 96   Version 1.1                                      */
/*                                                                           */
/* Function:                                                                 */
/*                                                                           */
/*   Procedure loads the specified range of positions within the given       */
/* sequence into a new string; returns the standard error structure, NULL if */
/* there were no errors in loading                                           */
/*                                                                           */
/*                                                                           */
/* Expected format of a sequence file:                                       */
/*                                                                           */
/*****************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ntl2.h"


/*****************************************************************************/

#define MNAME_SIZE                 128


/*****************************************************************************/

errind NTL2_LS_Get_Seq (char *file_name, char **seqptr, char **nameptr,
                     long int *fromptr, long int *toptr, long int *length);
errind extract_seq (char *file_name, char **seqptr,
                    long int from, long int to);
int ckgetc(FILE *fp);
errind NTL2_LS_Assemble_Error (int severity, int code, char *file_name,
                               char *comment, int description);

/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_Load_Sequence                                             */

errind NTL2_Load_Sequence (char *file_name, long int from, long int to,
                           char **into)
{
 char transfer [MNAME_SIZE], *str_var; long int int_1, int_2, length;
 errind err_stat;

 if ((from <= 0) || (from > to))
  return NTL2_LS_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE, file_name,
                                 "Improper position specification", 0);

 sprintf (transfer, "%s{%ld,%ld}", file_name, from, to);
 if ((err_stat = NTL2_LS_Get_Seq (transfer, into,
                                  &str_var, &int_1, &int_2, &length)) != NULL)
  return err_stat;
 else if (length != to - from + 1)
  return NTL2_LS_Assemble_Error (USER_ERROR, ERR_BAD_VALUE, file_name,
                          "Unmatching size returned from sequence loading", 0);
 else return NULL;
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_LS_Get_Seq                                                */

errind NTL2_LS_Get_Seq (char *file_name, char **seqptr, char **nameptr,
                        long int *fromptr, long int *toptr, long int *length)
{
 FILE *fp; char *p; int c; long int n;

 /* Check for position specification                                         */
 
 if ((p = strchr (file_name, '{')) != NULL) {
  sscanf (p+1, "%ld,%ld", fromptr, toptr);
  if ((*fromptr <= 0) || (*fromptr > *toptr))
   return NTL2_LS_Assemble_Error (USER_ERROR, ERR_BAD_VALUE, file_name,
                                  "Improper positions specification", 0);
  c = *p; *p = '\0'; *nameptr = NTL0_strsave(file_name);
  *p = c; *length = *toptr - *fromptr + 1;
  return extract_seq (*nameptr, seqptr, *fromptr, *toptr);
 }
 if ((fp = fopen (file_name, "r")) == NULL)
  return NTL2_LS_Assemble_Error (USER_ERROR, ERR_NO_VALUE, file_name,
                                 "Can't open the sequence file", 0);
 for (n = 0; (c = getc(fp)) != EOF; ) { if (c != '\n') ++n; }
 *seqptr = NTL0_ckalloc ((n+1) * sizeof(char));
 if (n == 0) { *length = 0; fclose(fp); return NULL; }
  
 rewind(fp); for (p = *seqptr; ; ) {
  if (fgets (p, (int) n, fp) == NULL) { fclose (fp); *p = '\0'; break; }
  if ((p == *seqptr) && (!isupper (*p))) continue;
  while (isupper(*p)) ++p;
  if ((*p != '\n') && (*p != '\0')) {
   fprintf(stderr, "Illegal character %c in file %s.\n", *p, file_name);
   return NTL2_LS_Assemble_Error (USER_ERROR, ERR_BAD_VALUE, file_name,
                                  "Illegal character in sequence", 0);
  }
 }
 *length = p - *seqptr;

 /* Defaults:                                                                */
  
 *nameptr = file_name; *fromptr = 1; *toptr = *length;
 return NULL;
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: extract_seq                                                    */

errind extract_seq (char *file_name, char **seqptr, long int from, long int to)
{
 FILE *fp; int c; long int i; char *p;

 *seqptr = (char *) NTL0_ckalloc ((to - from + 2) * sizeof(char));
 if ((fp = fopen (file_name, "r")) == NULL)
  return NTL2_LS_Assemble_Error (USER_ERROR, ERR_NO_VALUE, file_name,
                                 "Can't open the sequence file", 0);

 while (!isalpha (c = ckgetc(fp)))               /* Flush non-sequence lines */
  while (c != '\n') c = ckgetc(fp);
 ungetc(c, fp);

 for (i = 0; i < from-1; ) if (isalpha (ckgetc(fp))) ++i;
 for (p = *seqptr; i < to; ) if (isalpha (c = ckgetc(fp))) { *p++ = c; ++i; }
 *p = '\0'; fclose(fp); return NULL;
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: ckgetc                                                         */

int ckgetc (FILE *fp)
{
 int c;

 if ((c = getc (fp)) == EOF) {
  fprintf(stderr, "Unexpected end of sequence file.\n"); exit (1);
 }
 return c;
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_LS_Assemble_Error                                         */
/*                                                                           */
/* Service procedure for assembling and returnning an error report, based on */
/*   the values of the input parameters; returns the record with the report  */

errind NTL2_LS_Assemble_Error (int severity, int code, char *file_name,
                               char *comment, int description)
{
 char *report; errind assembled;

 report = (char *) NTL0_ckalloc (
                     (strlen (comment) + 2 * strlen ("_Load_Sequence: ") +
                      strlen (file_name) + strlen ("Sequence file: ") + 2) *
                                                                sizeof (char));
 sprintf (report, "_Load_Sequence: %s\n                Sequence file: %s",
          comment, file_name);
 assembled = NTL1_Error_Record (severity, code, report, description);
 free (report); return assembled;
}


